package ru.dbotthepony.kstarbound.defs.item

import com.google.gson.Gson
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kstarbound.defs.image.AtlasConfiguration
import ru.dbotthepony.kstarbound.defs.image.ImageReference
import ru.dbotthepony.kstarbound.io.json.builder.FactoryAdapter
import ru.dbotthepony.kstarbound.io.json.builder.JsonImplementation
import ru.dbotthepony.kstarbound.util.AssetPathStack

interface IArmorItemDefinition : ILeveledItemDefinition, IScriptableItemDefinition {
	/**
	 * @see ArmorPieceType
	 */
	val armorType: ArmorPieceType

	/**
	 * Варианты покраски (???)
	 */
	val colorOptions: List<Map<String, String>>

	/**
	 * Визуальные кадры анимации, когда надето на гуманоида мужского пола
	 */
	val maleFrames: IArmorFrames

	/**
	 * Визуальные кадры анимации, когда надето на гуманоида женского пола
	 */
	val femaleFrames: IArmorFrames

	@JsonImplementation(ArmorFrames::class)
	interface IArmorFrames {
		val body: ImageReference
		val backSleeve: ImageReference?
		val frontSleeve: ImageReference?
	}

	data class ArmorFrames(
		override val body: ImageReference,
		override val backSleeve: ImageReference? = null,
		override val frontSleeve: ImageReference? = null,
	) : IArmorFrames {
		class Factory(private val remapper: AssetPathStack, private val atlasRegistry: () -> AtlasConfiguration.Registry) : TypeAdapterFactory {
			override fun <T : Any?> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
				if (type.rawType == ArmorFrames::class.java) {
					return object : TypeAdapter<ArmorFrames>() {
						private val adapter = FactoryAdapter.Builder(
							ArmorFrames::class,
							ArmorFrames::body,
							ArmorFrames::backSleeve,
							ArmorFrames::frontSleeve,
						).build(gson)

						override fun write(out: JsonWriter, value: ArmorFrames?) {
							if (value == null)
								out.nullValue()
							else
								adapter.write(out, value)
						}

						override fun read(`in`: JsonReader): ArmorFrames? {
							if (`in`.peek() == JsonToken.NULL)
								return null

							if (`in`.peek() == JsonToken.STRING) {
								val path = remapper.remap(`in`.nextString())
								return ArmorFrames(ImageReference(path, atlasRegistry.invoke().get(path)), null, null)
							}

							return adapter.read(`in`)
						}
					} as TypeAdapter<T>
				}

				return null
			}
		}
	}
}
